# Simple calculator where a 4x4 matrix keypad and
# an alphanumeric LCD is interfaced by an MCP23017
# GPIO Expander
#
# The concept and the technical details of handling
# the 4x4 key matrix is based on the
# Application Note AN1081a of Microchip Technology Inc.
# Reference: https://www.microchip.com/en-us/application-notes/an1081
#
# The LCD is driven by the Adafruit_CircuitPython_CharLCD library
# Reference: https://github.com/adafruit/Adafruit_CircuitPython_CharLCD
#
# The calculator is a very simple one based on the article:
# Basic calculator program using Python
# Reference: https://www.geeksforgeeks.org/make-simple-calculator-using-python/
#
# The MCP23017 is driven by the Adafruit_CircuitPython_MCP230xx library
# Reference: https://github.com/adafruit/Adafruit_CircuitPython_MCP230xx
#
# Tested with: STM32F411CE Black Pill with Flash and CircuitPython v7.3.3

import time
import board
import busio
from digitalio import *
from adafruit_mcp230xx.mcp23017 import MCP23017
import adafruit_character_lcd.character_lcd as characterlcd

decoder = {}
decoder[0x77] = '1'
decoder[0x7B] = '2'
decoder[0x7D] = '3'
decoder[0x7E] = '+'
decoder[0xB7] = '4'
decoder[0xBB] = '5'
decoder[0xBD] = '6'
decoder[0xBE] = '-'
decoder[0xD7] = '7'
decoder[0xDB] = '8'
decoder[0xDD] = '9'
decoder[0xDE] = '*'
decoder[0xE7] = 'C'
decoder[0xEB] = '0'
decoder[0xED] = '='
decoder[0xEE] = '/'

entry = "0"

# board.A1 accepts the interrupt signal of the MCP230xx
irq_a = DigitalInOut(board.A1)
irq_a.direction = Direction.INPUT
irq_a.pull = Pull.UP

# Initialize the I2C bus:
i2c = busio.I2C(board.SCL, board.SDA)
mcp = MCP23017(i2c, address=0x20, reset=True)    # MCP23017

# Modify this if you have a different sized character LCD
lcd_columns = 16
lcd_rows = 2

# LCD is connected to the GPIOB port of the MCP23019 IC:
lcd_rs = mcp.get_pin(8)    # B0
lcd_en = mcp.get_pin(9)    # B1
lcd_backlight = mcp.get_pin(10)   # B2
lcd_d4 = mcp.get_pin(12)   # B4
lcd_d5 = mcp.get_pin(13)   # B5
lcd_d6 = mcp.get_pin(14)   # B6
lcd_d7 = mcp.get_pin(15)   # B7

# Initialize the LCD class
lcd = characterlcd.Character_LCD_Mono(
    lcd_rs, lcd_en, lcd_d4, lcd_d5, lcd_d6, lcd_d7, lcd_columns, lcd_rows, lcd_backlight
)


def gpio_init():
    mcp.iocon = 0
    mcp.iodira = 0xF0
    mcp.gppua = 0xF0
    mcp.gpioa = 0
    mcp.interrupt_configuration = 0xF0
    mcp.default_value = 0xF0
    mcp.interrupt_enable = 0xF0

def key_scan()-> int:
    mcp.gpioa = 0
    r = (mcp.gpioa) & 0xF0 # Read row
    #--- Flip I/O ---------------------
    mcp.iodira = 0x0F
    mcp.gppua = 0x0F
    mcp.interrupt_configuration = 0x0F
    mcp.default_value = 0x0F
    mcp.interrupt_enable = 0x0F
    #--- Determine key code
    c = (mcp.gpioa) & 0x0F # Read column
    # print("Row: {0}, Column: {1}".format(hex(r>>4),hex(c)))
    #--- Re-initialize I/O ------------
    mcp.iodira = 0xF0
    mcp.gppua = 0xF0
    mcp.interrupt_configuration = 0xF0
    mcp.default_value = 0xF0
    mcp.interrupt_enable = 0xF0
    return (r | c)

def wait_for_release():
    while not irq_a.value:
        time.sleep(0.005)
        dummy = mcp.int_capa

lcd.backlight = True               # Turn backlight on
lcd.clear()                        # Clear screen
lcd.message = "cPy calculator\n0"
gpio_init()                        # Initialize keyboard

while True:
    if not irq_a.value:
        keycode = key_scan()
        if keycode in decoder:
            ch = decoder[keycode]
            if ch == 'C':
                entry = '0'
                lcd.clear()
                lcd.message = "cPy calculator\n0"
            elif ch == '=':
                try:
                    y = str(eval(entry))
                    lcd.clear()
                    lcd.message = entry + '\n=' + y
                    entry = y
                except SyntaxError:
                    lcd.clear()
                    lcd.message = entry + '\nSynatax error!'
                    entry ='0'
                except ZeroDivisionError:
                    lcd.clear()
                    lcd.message = entry + '\nDiv by zero!'
                    entry = '0'
            else:
                if entry == '0' and ch in '0123456789':
                    entry = ''
                    lcd.message = "cPy calculator\n"
                entry = entry + ch
                lcd.message = lcd.message + ch
        wait_for_release()